# Drupal on MongoDB



## Requirements

The full support database driver for MongoDB has the following minimum requirements:
 - Drupal 11.1
 - MongoDB 7.0 with a replica set. The replica set offers support for [multi document transactions](https://www.mongodb.com/docs/v5.2/core/transactions/).



## How is Drupal's entity data stored in MongoDB

All entity instances are stored in JSON documents.
Each entity instance is stored in its own JSON document.
All entity data including all revisional data, all translations and all field data belonging to an entity instance are stored in the same JSON document.
Because all entity data for an entity instance is stored in the same JSON document, loading an entity instance from a MongoDB database is always a single row from a single table. Or in MongoDB jargon: A single document from a single collection.
No more complicated join queries are required for loading an entity from the relational database (MySQL, MariaDB or PostgreSQL).

An example of how entity data is stored in MongoDB (user #1):
```
{
    _id: ObjectId('664afdd4a3a001e71e0b49c7'),
    uid: 1,
    uuid: '841149cd-fe56-47c4-a112-6d23f561332f',
    langcode: 'en',
    user_translations: [
        {
            uid: 1,
            uuid: '841149cd-fe56-47c4-a112-6d23f561332f',
            langcode: 'en',
            preferred_langcode: 'en',
            name: 'root',
            pass: '$2y$10$kjGuIsPOTDa2TseuWMFGS.veLzH/khl0SfsuZNAeRPRtABgfq5GSC',
            mail: 'admin@finalist.nl',
            timezone: 'Europe/Amsterdam',
            status: true,
            created: ISODate('2024-05-20T07:37:54.000Z'),
            changed: ISODate('2024-05-20T07:42:08.000Z'),
            access: ISODate('2024-05-20T08:46:47.000Z'),
            login: ISODate('2024-05-20T07:44:16.000Z'),
            init: 'admin@finalist.nl',
            default_langcode: true,
            user_translations__roles: [
                {
                    bundle: 'user',
                    deleted: false,
                    langcode: 'en',
                    entity_id: 1,
                    revision_id: 1,
                    delta: 0,
                    roles_target_id: 'administrator'
                }
            ]
        }
    ],
    login: ISODate('2024-05-20T07:44:16.000Z'),
    access: ISODate('2024-05-20T08:46:47.000Z')
}
```

 - Every JSON document in MongoDB has its own unique ID with its own key `_id`.
 - Timestamps are stored as ISODate objects. Timestamps are stored with their timezone. Timestamps can be printed in many ways including other timezones.
 - Boolean values are stored as real booleans, not as integers as is currently the case in Drupal.
 - The user entity and therefore all entity data that is translatable is stored in an embedded table. For the user entity the embedded table name is `user_translations`. In the example the root user has only a single translation. Every translation is given its own row in the embedded translations table. Revisional data is also stored in its own embedded tables. An entity that is revisionable has 3 embedded tables, one for the current revision, one for all revisions and one for the latest revision. This is needed to support entity queries.
 - Entity field data is stored in its own embedded table. Every field has its own embedded table. In the given example the root user has the role "administrator". Every role that the user is given its own row in the embedded table `user_translations__roles`.

When entity instances are revisionable, translatable and/or have field data attached to it, may result in big JSON documents. For instance when you install the Unami profile the node with the recipe 'Deep mediterranean quiche' has 2 revisions, 2 translations and 10 fields with data. When you view the JSON document for this node in the Mongo Express it is 1764 lines long. The size is about 65K bytes. The maximum size of a JSON document is 16M bytes. The maximum JSON document size is big, but not endless.



## Current status of the database driver

 - All the base functions of Drupal Core are functional.
 - Installing a Drupal site with Drupal UI process are functional. Over 90% of the Drupal Core kerneltests and functionaltests pass successfully.
 - Drupal's Database API is supported. Some complicated queries, like with subqueries are not supported.
 - Drupal's EntityStorage API is supported.
 - Drupal's EntityQuery API is supported with the exception of relationship queries.
 - Drupal's views module is supported with the exception of views with groupwise max functionality.
 - Bugs occur when using the modules content_moderation, workflows and workspaces.
 - The database driver is not tested with contrib or custom modules.



## Install Drupal on MongoDB (for Drupal Core [11.1.6](https://www.drupal.org/project/drupal/releases/11.1.6))

This install guide uses the DDEV development environment.

This install guide is based on the following DDEV [guide](https://ddev.readthedocs.io/en/stable/users/quickstart/#drupal).

### 1. Create a directory for the Drupal on MongoDB project.
```mkdir your-project-directory && cd your-project-directory```

### 2. Configure your DDEV project. Disable the default management settings and the default database. We do not need a MySQL database.
```ddev config --project-type=drupal --php-version=8.3 --docroot=web --omit-containers=db --disable-settings-management=true```

```ddev start```

```ddev composer create drupal/recommended-project:11.1.6```

```ddev config --update```

```ddev restart```

### 3. Install the MongoDB database with the replica set.
```ddev add-on get drupal-daffie/ddev-mongo-replicaset```

```ddev restart```


### 4. Add the MongoDB database driver module to the project
```ddev composer require drupal/mongodb:^3.1```

### 5. Drupal Core needs to be patched to make it all work.
```cd web```

```git apply -v modules/contrib/mongodb/patches/drupal-core-11.1.6.patch```

```cd ..```

### 6. Optional: Install Drush
```ddev composer require drush/drush```

### 7. Do a regular site install with the UI.

On the database configuration page select **MongoDB**

 - Database name: `db`
 - Database username: `db`
 - Database password: `db`
 - Database replica set: `dbrs`
 - Host: `mongo`

![Database options to set](database-settings-example2.png)

## The CI Pipeline

All the Drupal Core tests are run in (Drupal on MongoDB (the full PR) (11.1.x))[https://www.drupal.org/project/drupal/issues/3503946] with the regular CI Pipeline.

## Development

Development is being done in (Drupal on MongoDB (the full PR))[https://www.drupal.org/project/drupal/issues/3403695].

Any help will be very much appreciated, only please do not change my personal MR. Copy my MR and create your own.

## Connect to MongoDB with [Mongo Express](https://github.com/mongo-express/mongo-express)

Mongo Express is a web based admin interface to a MongoDB database. To log in use the username `db` and the password `db`.



## Connect to MongoDB with MongoDB's [Compass](https://www.mongodb.com/products/tools/compass) tool

MongoDB has a great FREE tool to connect to a MongoDB database.
> Compass is a free interactive tool for querying, optimizing, and analyzing your MongoDB data. Get key insights, drag and drop to build pipelines, and more.

The connection string that I use, and yes it can be improved, is:
```
mongodb://db:db@your-project-directory.ddev.site:27017/?directConnection=true&replicaSet=dbrs
```
